@@ -32,8 +32,10 @@
 #include <iostream>
 #include <unistd.h>
 
-#include "parser/mstring.h"
 #include "klsconfig.h"
+#include "parser/mstring.h"
+#include "parser/htmlparser.h"
+#include "interfaces/engine/isearchmanager.h"
 
 
 SearchManager::SearchManager(int max_simultaneous_connections, int time_out,
@@ -45,17 +47,24 @@ SearchManager::SearchManager(int max_simultaneous_connections, int time_out,
         finished_connections_(max_simultaneous_connections_),
         maximum_current_connections_(-1), general_domain_(false),
         checked_general_domain_(false), time_out_(time_out), current_connections_(0),
-        send_identification_(true), canceled_(false), searching_(false), checked_links_(0), ignored_links_(0),
+        send_identification_(true), canceled_(false), searching_(false),  ignored_links_(0),
         check_parent_dirs_(true), check_external_links_(true), check_regular_expressions_(false),
         number_of_current_level_links_(0), 
-        links_rechecked_(0), recheck_current_index_(0)
+        links_rechecked_(0), recheck_current_index_(0), m_weaver(this)
 {
     kDebug(23100) <<  "SearchManager::SearchManager()";
 
     root_.setIsRoot(true);
 
+    if (KLSConfig::userAgent().isEmpty()) {
+        KLSConfig::setUserAgent(KProtocolManager::defaultUserAgent());
+    }
+    user_agent_ = KLSConfig::userAgent();
+
     m_weaver.setMaximumNumberOfThreads(10);
     connect(&m_weaver, SIGNAL(jobDone(ThreadWeaver::Job*)), SLOT(slotJobDone(ThreadWeaver::Job*)));
+    
+    new ISearchManager(this);
 }
 
 void SearchManager::reset()
@@ -85,12 +94,12 @@ void SearchManager::reset()
     current_connections_ = 0;
     canceled_ = false;
     searching_ = false;
-    checked_links_ = 0;    
+    search_counters_.total_links_ = 0;
     if(KLSConfig::userAgent().isEmpty()) {
         KLSConfig::setUserAgent(KProtocolManager::defaultUserAgent());
     }
     user_agent_ = KLSConfig::userAgent();
-
+    
     removeHtmlParts();
 }
 
@@ -142,7 +151,7 @@ void SearchManager::recheckLinks(QList<LinkStatus*> const& linkstatus_list)
     kDebug(23100) << "SearchManager::recheckLinks: " << linkstatus_list.size();
   
     Q_ASSERT(!searching_);
-    Q_ASSERT(checked_links_ >= linkstatus_list.size());
+    Q_ASSERT(search_counters_.total_links_ >= linkstatus_list.size());
     Q_ASSERT(search_results_.size() != 0);
 
     recheck_mode_ = true;
@@ -230,10 +239,17 @@ void SearchManager::finnish()
     }
     kDebug(23100) << "SearchManager::finnish";
     if(!recheck_mode_)
-        kDebug(23100) << "Links Checked: " << checked_links_;
+        kDebug(23100) << "Links Checked: " << search_counters_.total_links_;
     else
         kDebug(23100) << "Links Rechecked: " << links_rechecked_;
 
+    // Delete thread jobs
+//     QList<Job*> jobs = findChildren<Job*> ();
+//     foreach(Job* job, jobs) {
+//         job->setParent(0);
+//         job->deleteLater();
+//     }
+    
     searching_ = false;
     emit signalSearchFinished(this);
 }
@@ -275,14 +291,15 @@ void SearchManager::slotRootChecked(LinkStatus* link, LinkChecker* checker)
     kDebug(23100) <<  link->absoluteUrl().url() << " -> " << 
             LinkStatusHelper::lastRedirection(&root_)->absoluteUrl().url() << endl;
 
-    Q_ASSERT(checked_links_ == 0);
+    Q_ASSERT(search_counters_.total_links_ == 0);
     Q_ASSERT(search_results_.size() == 0);
 
     if(KLSConfig::showMarkupStatus() && link->isHtmlDocument())
         LinkStatusHelper::validateMarkup(link);
     
-    ++checked_links_;
-    //kDebug(23100) <<  "++checked_links_: SearchManager::slotRootChecked";
+    ++search_counters_.total_links_;
+    search_counters_.updateCounters(link);
+
     emit signalRootChecked(link);
     if(link->isRedirection() && link->redirection())
         linkRedirectionChecked(link->redirection());
@@ -294,7 +311,7 @@ void SearchManager::slotRootChecked(LinkStatus* link, LinkChecker* checker)
 
         QList<LinkStatus*> node;
         fillWithChildren(LinkStatusHelper::lastRedirection(&root_), node);
-
+        
         emit signalLinksToCheckTotalSteps(node.size());
 
         QList< QList<LinkStatus*> > nivel;
@@ -332,12 +349,24 @@ void SearchManager::slotRootChecked(LinkStatus* link, LinkChecker* checker)
 
 void SearchManager::fillWithChildren(LinkStatus* link, QList<LinkStatus*>& children)
 {
+//     kDebug(23100) << "SearchManager::fillWithChildren - " << link->absoluteUrl().url();
+  
     if(!link || link->absoluteUrl().hasRef())
         return;
 
-    QList<Node*> const& nodes = link->childrenNodes();
-//     children.reserve(nodes.size());
+    HtmlParser parser(link->docHtml());
 
+    if(parser.hasBaseUrl())
+        link->setBaseURI(KUrl(parser.baseUrl().url()));
+    if(parser.hasTitle())
+        link->setHtmlDocTitle(parser.title().attributeTITLE());
+
+    link->setChildrenNodes(parser.anchorNodes());
+    // Free some space
+    link->setDocHtml(QString());
+
+    QList<Node*> const& nodes = parser.nodes();
+  
     QHash<KUrl, LinkStatus*> children_hash;
     children_hash.reserve(nodes.size());
     
@@ -399,9 +428,14 @@ void SearchManager::fillWithChildren(LinkStatus* link, QList<LinkStatus*>& child
 
 //             kDebug(23100) << "Going in: " << url << endl;
         }
+        else {
+            delete node;
+        }
     }
             
     search_results_hash_ = search_results_hash_.unite(children_hash);
+
+//     kDebug(23100) << "SearchManager::fillWithChildren - end";
 }
 
 bool SearchManager::existUrl(KUrl const& url, KUrl const& url_parent) const
@@ -498,7 +532,7 @@ void SearchManager::continueSearch()
                 emit signalAddingLevel(true);
 
 //                 kDebug(23100) << "ThreadWeaver Queue length: " << m_weaver.queueLength();
-                m_weaver.enqueue(new AddLevelJob(*this));
+                m_weaver.enqueue(new AddLevelJob(this));
             }
             else
             {
@@ -573,6 +607,7 @@ void SearchManager::checkLinksSimultaneously(QList<LinkStatus*> const& links, bo
 
 void SearchManager::checkLink(LinkStatus* ls, bool recheck)
 {
+//     kDebug(23100) << "SearchManager::checkLink - " << ls->absoluteUrl();
     Q_ASSERT(ls);
 
     QString protocol = ls->absoluteUrl().protocol();
@@ -618,30 +653,34 @@ void SearchManager::checkLink(LinkStatus* ls, bool recheck)
 
 void SearchManager::recheckLink(LinkStatus* ls)
 {
+    
+  
     checkLink(ls, true);
 }
 
 void SearchManager::linkRedirectionChecked(LinkStatus* link, bool recheck)
 {
-    kDebug(23100) <<  "SearchManager::linkRedirectionChecked: " << checked_links_;
+    kDebug(23100) <<  "SearchManager::linkRedirectionChecked: " << search_counters_.total_links_;
     
     emit signalRedirection();
     recheck ? emit signalLinkRechecked(link) : emit signalLinkChecked(link);
   
     if(!recheck) {
-        ++checked_links_;
+        ++search_counters_.total_links_;
         search_results_hash_.insert(link->absoluteUrl(), link);
     } else {
         ++links_rechecked_;
     }
     
+    search_counters_.updateCounters(link);
+    
     if(link->isRedirection() && link->redirection())
         linkRedirectionChecked(link->redirection(), recheck);
 }
 
 void SearchManager::slotLinkChecked(LinkStatus* link, LinkChecker* checker)
 {
-    kDebug(23100) <<  "SearchManager::slotLinkChecked: " << checked_links_;
+    kDebug(23100) <<  "SearchManager::slotLinkChecked: " << search_counters_.total_links_;
 //     kDebug(23100) <<  link->absoluteUrl().url() << " -> " << 
 //             LinkStatus::lastRedirection((const_cast<LinkStatus*> (link)))->absoluteUrl().url() << endl;
 
@@ -650,21 +689,23 @@ void SearchManager::slotLinkChecked(LinkStatus* link, LinkChecker* checker)
     Q_ASSERT(link);
 
     if(KLSConfig::showMarkupStatus() && link->isHtmlDocument())
-      LinkStatusHelper::validateMarkup(link);
+        LinkStatusHelper::validateMarkup(link);
 
     emit signalLinkChecked(link);
     
     if(link->isRedirection() && link->redirection())
         linkRedirectionChecked(link->redirection());
     
-    ++checked_links_;
+    ++search_counters_.total_links_;
     ++finished_connections_;
     --links_being_checked_;
+    
+    search_counters_.updateCounters(link);
   
     Q_ASSERT(links_being_checked_ >= 0);
 
     if(search_mode_ != depth || current_depth_ < depth_) {
-        m_weaver.enqueue(new BuildNodeJob(*this, link));
+        m_weaver.enqueue(new BuildNodeJob(this, link));
     }
 
     if(canceled_ && searching_ && !links_being_checked_)
@@ -693,6 +734,8 @@ void SearchManager::slotLinkRechecked(LinkStatus* link, LinkChecker* checker)
     
     Q_ASSERT(link);
 
+    search_counters_.updateCounters(link);
+
     if(KLSConfig::showMarkupStatus() && link->isHtmlDocument())
         LinkStatusHelper::validateMarkup(link);
 
@@ -731,7 +774,7 @@ void SearchManager::buildNewNode(LinkStatus* linkstatus)
 
     QList<LinkStatus*> new_node;
     fillWithChildren(linkstatus, new_node);
-    
+
     if(new_node.size() == 0)
         return;
         
@@ -877,11 +920,13 @@ void SearchManager::addHtmlPart(QString const& key_url, KHTMLPart* html_part)
 //     if(html_parts_.size() > 300)
 //         removeHtmlParts();
 
-    html_parts_.insert(key_url, html_part);
+//     html_parts_.insert(key_url, html_part);
 }
 
 void SearchManager::removeHtmlParts()
 {
+    kDebug() << "SearchManager::removeHtmlParts";
+    
     for(KHTMLPartMap::iterator it = html_parts_.begin(); it != html_parts_.end(); ++it)
     {
         delete it.value();
@@ -933,6 +978,22 @@ void SearchManager::save(QDomElement& element, LinkStatusHelper::Status status)
     child_element = element.ownerDocument().createElement("link_list");
     element.appendChild(child_element);
     
+    // root
+    if(root_.checked() && LinkStatusHelper::hasStatus(&root_, status)) {
+        LinkStatusHelper::save(&root_, child_element);
+    }
+    // redirections
+    LinkStatus const* ls = &root_;
+    while(ls->isRedirection()) {
+        ls = ls->redirection();
+        if(!ls) {
+            break;
+        }
+        if(ls->checked() && LinkStatusHelper::hasStatus(ls, status)) {
+            LinkStatusHelper::save(ls, child_element);
+        }
+    }
+
     for(int i = 0; i != search_results_.size(); ++i)
     {
         for(int j = 0; j != search_results_[i].size() ; ++j)
@@ -940,14 +1001,27 @@ void SearchManager::save(QDomElement& element, LinkStatusHelper::Status status)
             for(int l = 0; l != (search_results_[i])[j].size(); ++l)
             {
                 LinkStatus* ls = ((search_results_[i])[j])[l];
-                if(ls->checked() && LinkStatusHelper::hasStatus(ls, status))
+
+                if(ls->checked() && LinkStatusHelper::hasStatus(ls, status)) {
                     LinkStatusHelper::save(ls, child_element);
+                }
+                                    
+                while(ls->isRedirection()) {
+                    ls = ls->redirection();
+                    if(!ls) {
+                        break;
+                    }
+
+                    if(ls->checked() && LinkStatusHelper::hasStatus(ls, status)) {
+                        LinkStatusHelper::save(ls, child_element);
+                    }
+                }
             }
         }
     } 
 }
 
-QString SearchManager::toXML() const
+QString SearchManager::toXML(LinkStatusHelper::Status status) const
 {
     QDomDocument doc;
     doc.appendChild(doc.createProcessingInstruction( "xml", 
@@ -956,7 +1030,7 @@ QString SearchManager::toXML() const
     QDomElement root = doc.createElement("klinkstatus");
     doc.appendChild(root);
 
-    save(root);
+    save(root, status);
     
     return doc.toString(4);
 }
@@ -968,7 +1042,8 @@ void SearchManager::slotJobDone(ThreadWeaver::Job* job)
         slotLevelAdded();
     }
 
-    delete job;
+//     delete job;
+    job->deleteLater();
 }
 
 QStringList SearchManager::findUnreferredDocuments(KUrl const& baseDir, QStringList const& documentList) const
@@ -1042,41 +1117,83 @@ QList<LinkStatus*> SearchManager::getLinksWithHtmlProblems() const
 
 // BuildNodeJob
 
-BuildNodeJob::BuildNodeJob(SearchManager& manager, LinkStatus* linkstatus)
-  : m_searchManager(manager), m_linkStatus(linkstatus)
+BuildNodeJob::BuildNodeJob(SearchManager* manager, LinkStatus* linkstatus)
+  : Job(/*manager*/), m_searchManager(manager), m_linkStatus(linkstatus)
 {
 }
 
 BuildNodeJob::~BuildNodeJob()
 {
+//     kDebug(23100) << "BuildNodeJob::~BuildNodeJob()";
 }
     
 void BuildNodeJob::run()
 {
 //     kDebug(23100) << "\n\n\nBuildNodeJob::run\n\n\n";
-    m_searchManager.buildNewNode(m_linkStatus);
+    if (m_searchManager->canceled_) {
+        return;
+    }
+
+    m_searchManager->buildNewNode(m_linkStatus);
 }
 
 
 // AddLevelJob
 
-AddLevelJob::AddLevelJob(SearchManager& manager)
-  : m_searchManager(manager)
+AddLevelJob::AddLevelJob(SearchManager* manager)
+  : Job(/*manager*/), m_searchManager(manager)
 {
 }
 
 AddLevelJob::~AddLevelJob()
 {
+//     kDebug(23100) << "AddLevelJob::~AddLevelJob()";
 }
     
 void AddLevelJob::run()
 {
 //     kDebug(23100) << "\n\n\nAddLevelJob::run\n\n\n";
-    while(m_searchManager.m_weaver.queueLength() != 0) {
+    while(m_searchManager->m_weaver.queueLength() != 0) {
         kDebug(23100) << "AddLevelJob::run: waiting for running jobs to finish";
         sleep(1);
     }
-    m_searchManager.addLevel();
+ 
+    if (m_searchManager->canceled_) {
+        return;
+    } 
+
+    m_searchManager->addLevel();
 }
 
+
+SearchCounters::SearchCounters()
+  : total_links_(0), broken_links_(0), undetermined_links_(0)
+{
+}
+
+void SearchCounters::updateCounters(LinkStatus* link)
+{
+    if(LinkStatusHelper::hasStatus(link, LinkStatusHelper::bad)) {
+        ++broken_links_;
+    }
+    else if(LinkStatusHelper::hasStatus(link, LinkStatusHelper::undetermined)) {
+        ++undetermined_links_;
+    }
+}
+
+void SearchCounters::resetCounters(LinkStatus* link)
+{
+    if(link->isRedirection() && link->redirection()) {
+        resetCounters(link->redirection());
+    }
+  
+    if(LinkStatusHelper::hasStatus(link, LinkStatusHelper::bad)) {
+        --broken_links_;
+    }
+    else if(LinkStatusHelper::hasStatus(link, LinkStatusHelper::undetermined)) {
+        --undetermined_links_;
+    }
+}
+
+    
 #include "searchmanager.moc"
